@@ -258,7 +258,7 @@ _cogl_delete_gl_texture (GLuint gl_texture)
 /* Whenever the underlying GL texture storage of a CoglTexture is
  * changed (e.g. due to migration out of a texture atlas) then we are
  * notified. This lets us ensure that we reflush that texture's state
- * if it reused again with the same texture unit.
+ * if it is reused again with the same texture unit.
  */
 void
 _cogl_material_texture_storage_change_notify (CoglHandle texture)
@@ -273,7 +273,7 @@ _cogl_material_texture_storage_change_notify (CoglHandle texture)
         &g_array_index (ctx->texture_units, CoglTextureUnit, i);
 
       if (unit->layer &&
-          unit->layer->texture == texture)
+          _cogl_material_layer_get_texture (unit->layer) == texture)
         unit->texture_storage_changed = TRUE;
 
       /* NB: the texture may be bound to multiple texture units so
@@ -737,7 +737,9 @@ flush_layers_common_gl_state_cb (CoglMaterialLayer *layer, void *user_data)
        */
       if (unit->gl_texture != gl_texture || unit->is_foreign)
         {
-          if (unit_index != 1)
+          if (unit_index == 1)
+            unit->dirty_gl_texture = TRUE;
+          else
             GE (glBindTexture (gl_target, gl_texture));
           unit->gl_texture = gl_texture;
         }
@@ -831,9 +833,9 @@ _cogl_material_flush_common_gl_state (CoglMaterial  *material,
 
   state.i = 0;
   state.layer_differences = layer_differences;
-  _cogl_material_foreach_layer (material,
-                                flush_layers_common_gl_state_cb,
-                                &state);
+  _cogl_material_foreach_layer_internal (material,
+                                         flush_layers_common_gl_state_cb,
+                                         &state);
 
   /* Disable additional texture units that may have previously been in use.. */
   for (; state.i < ctx->texture_units->len; state.i++)
@@ -1113,9 +1115,9 @@ _cogl_material_flush_gl_state (CoglMaterial *material,
       memset (layer_differences, 0, sizeof (layer_differences));
       state.i = 0;
       state.layer_differences = layer_differences;
-      _cogl_material_foreach_layer (material,
-                                    compare_layer_differences_cb,
-                                    &state);
+      _cogl_material_foreach_layer_internal (material,
+                                             compare_layer_differences_cb,
+                                             &state);
     }
 
   /* First flush everything that's the same regardless of which
@@ -1176,9 +1178,9 @@ _cogl_material_flush_gl_state (CoglMaterial *material,
       state.layer_differences = layer_differences;
       state.error_adding_layer = FALSE;
       state.added_layer = FALSE;
-      _cogl_material_foreach_layer (material,
-                                    backend_add_layer_cb,
-                                    &state);
+      _cogl_material_foreach_layer_internal (material,
+                                             backend_add_layer_cb,
+                                             &state);
 
       if (G_UNLIKELY (state.error_adding_layer))
         continue;
